#include<netinet/in.h>
#include<sys/socket.h>
#include<net/gen/in.h>
#include"assert.h"
#include "const.h" 
#include"auth.h"
#include "xdr.h"
#include"xprt.h"
#include"uio.h"
#include "svcsock.h"
#include "rpc_msg.h"
#include "portmap.h"
#define OK 0
#define SERVICE_NOTFOUND -1

typedef struct pmaplist pmaplist;
typedef struct mapping mapping;
typedef unsigned int uint;
pmaplist * head;

pmaplist * getnode()
{
	pmaplist * node = (pmaplist *)malloc(sizeof(struct pmaplist));
	if(NULL == node) 
	{
		printf("\n Could not allocate memory for the pmaplist node ");
		exit(0);
	}
	node->next = NULL;
}
/* returns -1 incase service is not found */ 

uint searchPort(uint prog, uint vers, uint prot)
{
	pmaplist * node = head;
	uint port = SERVICE_NOTFOUND;
	while(node->next)
	{
		node = node->next; 
		if(prog == node->map.prog)
		{
			if(prot == node->map.prot)
			{
				if(vers == node->map.vers)
				{
					port = node->map.port;
					break;
				}
			}
		}
	} 	
	return port;
}

void addNode(uint prog, uint vers, uint prot, uint port)
{
	pmaplist * node;
	node = getnode();
	node->map.prog = prog;
	node->map.vers = vers;
	node->map.prot = prot;
	node->map.port = port;	
	/* Add it immediately after head */
	node->next = head->next;
	head->next = node;
	printf("\n Node added ! \n");
} 

int registerService(unsigned int prog, unsigned int vers, unsigned int prot, unsigned int port)
{
	int retVal = SERVICE_NOTFOUND ;
	retVal = searchPort(prog, vers, prot);
	if(SERVICE_NOTFOUND == retVal)
	{
		addNode(prog, vers, prot, port);
		retVal = 0;
	}
	else
	{
		/* service is found so trying to re-register */
		retVal = -1;
	}
	return retVal;
}

int delNode(uint prog, uint vers, uint prot)
{
	int retVal = SERVICE_NOTFOUND;
	pmaplist * node = head;
	pmaplist * temp;
	while(node->next)
	{
		temp = node->next;
		if(prog == temp->map.prog)
		{
			if(vers == temp->map.vers)
			{
				if(prot == temp->map.prot)
				{
					node->next = temp->next;
					free(temp);
					retVal = 0; 	
					break;	
				}	
			}	
		}	
	}
	return retVal;	
}	

int unregisterService(unsigned int prog, unsigned int vers, unsigned int prot, unsigned int port)
{
	int retVal = SERVICE_NOTFOUND ;
        retVal = delNode(prog, vers, prot);
	return retVal;
}	


void printXDR(char * buff, int len)
{
        int i = 0;
	printf("\n ********************************** \n");
        for(i=0; i<len; i++)
        {
                printf("\n %u ", ntohl(*(u_long *)buff));
                buff = buff + 4;
                len = len - 4;
        }
	printf("\n ********************************** \n");
}





void decodeMsg(char * buff, int sockFd, struct sockaddr_in sin)
{
	/*
	struct rpc_msg * cmsg = (struct rpc_msg *)buff;
	unsigned long xid = (int32_t)ntohl((uint32_t)(cmsg->rm_xid));
	u_long cb_rpcvers = (u_long)ntohl((u_long)(cmsg->ru.RM_cmb.cb_rpcvers));
	u_long cb_prog = (u_long)ntohl ( (u_long)(cmsg->ru.RM_cmb.cb_prog)); 
	u_long cb_vers = (u_long)ntohl((u_long)(cmsg->ru.RM_cmb.cb_vers));
	u_long cb_proc = (u_long)ntohl((u_long)(cmsg->ru.RM_cmb.cb_proc));
	printf("\n The transaction id received is: %d " , xid);
	printf("\n the version number is : %d", cb_rpcvers);
	printf("\n the cb_prog is: %d ", cb_prog);
	printf("\n the cb_proc is: %d ", cb_proc);
	printf("\n the cb_vers is : %d", cb_vers);
	printf("\n xid in buff is : %d " , (*(long *) buff));*/
	printXDR(buff, sizeof(struct rpc_msg));
	switch(cb_proc)
	{
		case PMAPPROC_SET:
			printf("\n the cb_proc is: PMAPPROC_SET" );
			registerService(prog, vers, prot, port);
			sendReply(buff, sockFd, sin, TRUE);
			break;
		case PMAPPROC_UNSET:
			printf("\n the cb_proc is: PMAPPROC_UNSET" );
			sendReply(buff, sockFd, sin, TRUE);
			break;			
		case PMAPPROC_GETPORT:
			printf("\n the cb_proc is: PMAPPROC_GETPROC" );
			sendReply(buff, sockFd, sin, 2049);
			break;

	}
	printf("\n");
}
#define UDPMSGSIZE      8800    /* rpc imposed limit on udp msg size */


int sendReply(char * buff, int sd, struct sockaddr_in sin, int result)
{	
	xdrproc_t xdr_int;
	XDR xdr;	
	int xid, slen;
	int port = 2049; /* for nfsd as of now ! */
	int sent;
	int len = sizeof(struct sockaddr_in);
	xdr.x_base = (char *)malloc( UDPMSGSIZE);
	xdr.x_private = xdr.x_base; 
	xid = (u_long)ntohl((uint32_t)((struct rpc_msg*) buff)->rm_xid);
	* (uint32_t *) xdr.x_private = htonl(xid);
	xdr.x_private = xdr.x_private + 4;
	*(int *)xdr.x_private =  htonl(REPLY);
	xdr.x_private = xdr.x_private + 4;
	*(int *)xdr.x_private =  htonl(MSG_ACCEPTED);
	xdr.x_private = xdr.x_private + 4;
	*(int *)xdr.x_private =  htonl(0); /* opaque_auth.oa_flavour*/
	xdr.x_private = xdr.x_private + 4;
	*(int *)xdr.x_private =  htonl(0); /* opaque_auth.oa_length */
	xdr.x_private = xdr.x_private + 4;
	 *(int *)xdr.x_private =  htonl(SUCCESS);
	xdr.x_private = xdr.x_private + 4; 
	*(int*)xdr.x_private = htonl(result); 
	xdr.x_private = xdr.x_private + 4;
	slen = xdr.x_private - xdr.x_base;
	sent = sendto(sd, xdr.x_base, slen, 0, (struct sockaddr *)&sin, len);		
	printf("\n %d ", htonl(2049));
	printf("\n slen is %d and Sent %d bytes \n", slen, sent);
	return (0);
}	

int portmap_process(int sockFd)
{
	int retVal = OK;
	/* as of now we know that the protocol being used is UDP */
	struct sockaddr_in sin;
	char buff[8800];
	int len;
	socklen_t lenAddr;
	printf("\n Inside portmap_process() \n");
	while(1)
	{
		len = recvfrom(sockFd, buff, 4096, 0, (struct sockaddr *) (&sin), &lenAddr);
		printf("\n Mesg recvd \n");
		if(len != -1)
		{
			decodeMsg(buff, sockFd, sin);
		} 
		else
		{
			printf("\n Could not receive anything !");
		}
	}
	return retVal;
}

static int fatal(char * str)
{
        printf("\n %s", str);
        exit(0);
}

int initPortmapDaemon(int portNum)
{
	struct svc_serv * serv;
	int protocol = IPPROTO_UDP;
/*  you store a socket descriptor in the svc_serv   */
        int sd, domain, type, retVal = 0;
        struct sockaddr_in sin;
        socklen_t address_len;
        assert(0>portNum);
        sin.sin_family = AF_INET;
        sin.sin_port = htons(portNum);
        sin.sin_addr.s_addr = INADDR_ANY;
        domain = AF_INET;
        type = (protocol == IPPROTO_UDP)? SOCK_DGRAM : SOCK_STREAM;
        sd = socket(domain, type, protocol);
        if(!sd)
        {
                fatal("Could not get any socket ! So aborting !");
        }
        if(0 > bind(sd,(struct sockaddr *) &sin, sizeof(sin)))
        {
                fatal("bind() from svc_setup_socket");
        }
        if(type == SOCK_STREAM)
        {
                if(0> (retVal = listen(sd,5)))
                {
                        fatal("listen() from svc_setup_socket");
                }
                printf("\n Listening on port %d\n", portNum);
        }
	head = getnode();
	head->map.prog = -1;
	head->map.vers = -1;
	head->map.prot = -1;
	head->map.port = -1; 
	printf("Socket created !");
        return sd;
}


int main(int arg, char * argv[])
{
	int sockfd;
	sockfd = initPortmapDaemon(111);
	portmap_process(sockfd);
}
